// http://mathr.co.uk/blog/2016-02-25_automated_julia_morphing.html

/*
adapt $prefix to where you installed mandelbrot-{numerics,symbolics}

$ gcc -std=c99 -Wall -Wextra -pedantic \
  -I$prefix/include/ -I../lib \
  -L$prefix/lib/ -L../lib \
  -o m-automorpth m-automorph.c \
  -lmpc -lmpfr -lgmp -lm -lperturbator \
  -lmandelbrot-numerics -lmandelbrot-symbolics \
  -lGLEW -lGL -lglfw -pthread -fopenmp -lstdc++ \
  -O3 -march=native
*/

#define _POSIX_C_SOURCE 199309L

#include <math.h>
#include <assert.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <mpc.h>
#include <mpfr.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <mandelbrot-numerics.h>
#include <mandelbrot-symbolics.h>
#include <perturbator.h>

static inline int max(int a, int b) {
  return a > b ? a : b;
}

static const char *simple_vert =
  "#version 130\n"
  "out vec2 texCoord;\n"
  "const vec2 t[4] = vec2[4](vec2(0.0, 0.0), vec2(1.0, 0.0), vec2(0.0, 1.0), vec2(1.0, 1.0));\n"
  "void main() {\n"
  "  gl_Position = vec4(t[gl_VertexID] * 2.0 - 1.0, 0.0, 1.0);\n"
  "  texCoord = t[gl_VertexID] * vec2(1.0, -1.0) + vec2(0.0, 1.0);\n"
  "}\n"
  ;

static const char *simple_frag =
  "#version 130\n"
  "uniform sampler2D tex;\n"
  "uniform bool show_lines;\n"
  "uniform float weight;\n"
  "in vec2 texCoord;\n"
  "out vec4 colour;\n"
  "int px(vec2 tc) {\n"
  "  vec4 t = texture(tex, tc);\n"
  "  int k = 0;\n"
  "  k += (t.x / 2.0) - floor(t.x / 2.0) >= 0.5 ? 1 : 2;\n"
  "  k += t.z >= 0.0 ? 4 : 8;\n"
  "  k += 16 * int(floor(t.x));\n"
  "  return k;\n"
  "}\n"
  "void main() {\n"
  "  float e = 1.0;\n"
  "  vec2 dx = dFdx(texCoord);\n"
  "  vec2 dy = dFdy(texCoord);\n"
  "  if (show_lines) {\n"
  "    e = px(texCoord) == px(texCoord + dx + dy) && px(texCoord + dx) == px(texCoord + dy) ? 1.0 : 0.0;\n"
  "  }\n"
  "  vec2 me = texture(tex, texCoord).xy;\n"
  "  float s = tanh(clamp(texture(tex, texCoord).w / weight, 0.0, 8.0));\n"
  "  colour = vec4(dot(me, me) <= 0.0 ? vec3(1.0, 0.7, 0.0) : vec3(mix(0.0, mix(0.9, 1.0, e), s)), 1.0);\n"
  "}\n"
  ;

struct state_s {
  GLFWwindow *window;

  bool show_lines;
  GLuint show_lines_u;
  double log2weight;
  GLuint weight_u;

  int width;
  int height;

  int precision;
  mpfr_t centerx;
  mpfr_t centery;
  mpfr_t radius;

};
typedef struct state_s state_t;

static void refresh_callback(void *user_pointer) {
  state_t *state = user_pointer;
  assert(state);
  glUniform1i(state->show_lines_u, state->show_lines);
  glUniform1f(state->weight_u, exp2f(state->log2weight));
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  glfwSwapBuffers(state->window);
}

void debug_program(GLuint program, const char *name) {
  if (program) {
    GLint linked = GL_FALSE;
    glGetProgramiv(program, GL_LINK_STATUS, &linked);
    if (linked != GL_TRUE) {
      fprintf(stderr, "%s: OpenGL shader program link failed\n", name);
    }
    GLint length = 0;
    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &length);
    char *buffer = (char *) malloc(length + 1);
    glGetProgramInfoLog(program, length, NULL, buffer);
    buffer[length] = 0;
    if (buffer[0]) {
      fprintf(stderr, "%s: OpenGL shader program info log\n", name);
      fprintf(stderr, "%s\n", buffer);
    }
    free(buffer);
  } else {
    fprintf(stderr, "%s: OpenGL shader program creation failed\n", name);
  }
}

void debug_shader(GLuint shader, GLenum type, const char *name) {
  const char *tname = 0;
  switch (type) {
    case GL_VERTEX_SHADER:   tname = "vertex";   break;
    case GL_GEOMETRY_SHADER: tname = "geometry"; break;
    case GL_FRAGMENT_SHADER: tname = "fragment"; break;
    default:                 tname = "unknown";  break;
  }
  if (shader) {
    GLint compiled = GL_FALSE;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);
    if (compiled != GL_TRUE) {
      fprintf(stderr, "%s: OpenGL %s shader compile failed\n", name, tname);
    }
    GLint length = 0;
    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
    char *buffer = (char *) malloc(length + 1);
    glGetShaderInfoLog(shader, length, NULL, buffer);
    buffer[length] = 0;
    if (buffer[0]) {
      fprintf(stderr, "%s: OpenGL %s shader info log\n", name, tname);
      fprintf(stderr, "%s\n", buffer);
    }
    free(buffer);
  } else {
    fprintf(stderr, "%s: OpenGL %s shader creation failed\n", name, tname);
  }
}

void compile_shader(GLint program, GLenum type, const char *name, const GLchar *source) {
  GLuint shader = glCreateShader(type);
  glShaderSource(shader, 1, &source, NULL);
  glCompileShader(shader);
  debug_shader(shader, type, name);
  glAttachShader(program, shader);
  glDeleteShader(shader);
}

GLint compile_program(const char *name, const GLchar *vert, const GLchar *frag) {
  GLint program = glCreateProgram();
  if (vert) { compile_shader(program, GL_VERTEX_SHADER  , name, vert); }
  if (frag) { compile_shader(program, GL_FRAGMENT_SHADER, name, frag); }
  glLinkProgram(program);
  debug_program(program, name);
  return program;
}

extern int main(int argc, char **argv) {
  state_t state;
  memset(&state, 0, sizeof(state));

  int workers = 4;
  int width = 1920;
  int height = 1080;
  int maxiters = 1 << 24;
  int chunk = 1 << 10;
  double escape_radius = 25;
  double glitch_threshold = 1e-6;
  double weight = 0;
  int sharpness = 8;

  m_block a, b;
  m_block_init(&a);
  m_block_init(&b);
  m_block_from_string(&a, "01111");
  m_block_from_string(&b, "10000");
  m_binangle as[2], bs[2];
  m_binangle_init(&as[0]);
  m_binangle_init(&bs[0]);
  m_binangle_from_string(&as[0], ".(011111000001111100000111110000011111000001111011111000011111000010000)");
  m_binangle_from_string(&bs[0], ".(011111000001111100000111110000011111000001111011111000100000111101111)");
  m_binangle_init(&as[1]);
  m_binangle_init(&bs[1]);

  mpq_t q;
  mpq_init(q);

  mpc_t c[2], delta;
  mpc_init2(c[0], 53);
  mpc_init2(c[1], 53);
  mpc_init2(delta, 53);

  mpfr_init2(state.radius, 53);
  mpfr_init2(state.centerx, 53);
  mpfr_init2(state.centery, 53);

  uint8_t *ppm = malloc(width * height * 3);
  assert(ppm);

  glfwInit();
  glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 1);
  glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
  GLFWwindow *window = glfwCreateWindow(width, height, "automorph", 0, 0);
  glfwMakeContextCurrent(window);
  glewExperimental = GL_TRUE;
  glewInit();
  glGetError(); // discard common error from glew

  GLuint tex;
  glGenTextures(1, &tex);
  glBindTexture(GL_TEXTURE_2D, tex);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, width, height, 0, GL_RGBA, GL_FLOAT, 0);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);

  GLuint vao;
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);

  GLuint program = compile_program("colourize", simple_vert, simple_frag);
  glUseProgram(program);
  state.show_lines_u = glGetUniformLocation(program, "show_lines");
  state.weight_u = glGetUniformLocation(program, "weight");

  state.window = window;
  state.width = width;
  state.height = height;
  state.log2weight = weight;

  struct perturbator *context = perturbator_new(workers, width, height, maxiters, chunk, escape_radius, glitch_threshold);

  glfwSetWindowUserPointer(window, &state);

  for (int depth = 0; depth < 20; ++depth) {
    glfwPollEvents();
    if (glfwWindowShouldClose(window)) {
      break;
    }
    int w = (depth + 1) & 1;
    m_binangle_to_rational(q, &as[1 - w]);
    m_r_exray_in *ray = m_r_exray_in_new(q, sharpness);
    for (int s = 0; s < 2 * as[1 - w].per.length * sharpness; ++s) {
      m_r_exray_in_step(ray);
    }
    m_r_exray_in_get(ray, c[1 - w]);
    m_r_nucleus(c[1 - w], c[1 - w], as[1 - w].per.length, 64);

    if (depth > 0) {
      mpc_sub(delta, c[w], c[1 - w], MPC_RNDNN);
      mpc_abs(state.radius, delta, MPFR_RNDN);
      mpfr_mul_d(state.radius, state.radius, 12, MPFR_RNDN);
      mpfr_set_prec(state.centerx, mpc_get_prec(c[w]));
      mpfr_set_prec(state.centery, mpc_get_prec(c[w]));
      mpfr_set(state.centerx, mpc_realref(c[w]), MPFR_RNDN);
      mpfr_set(state.centery, mpc_imagref(c[w]), MPFR_RNDN);
      perturbator_start(context, state.centerx, state.centery, state.radius);
      perturbator_stop(context, false);
      glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_FLOAT, perturbator_get_output(context));
      refresh_callback(&state);
      glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, ppm);
      printf("P6\n%d %d\n255\n", width, height);
      for (int y = height - 1; y >= 0; --y) {
        fwrite(ppm + y * width * 3, width * 3, 1, stdout);
      }
      fflush(stdout);
    }
    m_block_append(&as[w].per, &bs[1-w].per, &as[1-w].per);
    m_block_append(&as[w].per, &as[w].per, &b);
    m_block_append(&bs[w].per, &bs[1-w].per, &bs[1-w].per);
    m_block_append(&bs[w].per, &bs[w].per, &a);
  }

  m_binangle_clear(&as[0]);
  m_binangle_clear(&as[1]);
  m_binangle_clear(&bs[0]);
  m_binangle_clear(&bs[1]);
  m_block_clear(&a);
  m_block_clear(&b);
  mpq_clear(q);
  mpc_clear(delta);
  mpc_clear(c[0]);
  mpc_clear(c[1]);

  perturbator_stop(context, true);
  free(ppm);
  glfwDestroyWindow(window);
  glfwTerminate();
  return 0;
(void) argc;
(void) argv;
}
